      *  HASHR4: Create a hash of a stream file (IFS file).  The

      *          hash can be MD5, SHA1, SHA256, SHA384 or SHA512.

      *                                 Scott Klement, April 6, 2018

      *

      * To Compile:

      *      CRTBNDRPG HASHR4 SRCFILE(xxx/QRPGLESRC) DBGVIEW(*LIST)

      *      CRTCMD CMD(HASH) SRCFILE(xxx/QCMDSRC) PGM(xxx/HASHR4)

      *

      * Note: Because the first parameter to this program is larger

      *       than 32 bytes, you must call this through the

      *       command interface.

      *

     H DFTACTGRP(*NO) BNDDIR('QC2LE')



      /copy ifsio_h



     D HASHR4          PR                  ExtPgm('HASHR4')

     D  in_Stmf                    5000A   varying

     D  in_Alg                       10a   const

     D HASHR4          PI

     D  in_Stmf                    5000A   varying

     D  in_Alg                       10a   const



     D ERRC0100_t      ds                  qualified

     D                                     template

     D   bytesProv                   10i 0 inz(%size(ERRC0100_t))

     D   bytesAvail                  10i 0 inz(0)

     D   msgId                        7a

     D                                1a

     D   msgDta                   32767a



     D Qc3CreateAlgorithmContext...

     D                 PR                  ExtProc('Qc3CreateAlgorithmContext')

     D   algDesc                  32767a   const options(*varsize)

     D   algDescFmt                   8a   const

     D   ctxToken                     8a

     D   errorCode                         likeds(ERRC0100_t)



     D ALGD0100        ds                  qualified

     D    ctxToken                    8a   inz(*ALLx'00')

     D    finalFlag                   1a   inz(*OFF)



     D ALGD0500        ds                  qualified

     D    hashAlg                    10i 0



     D HASH_MD5        c                   const(1)

     D HASH_SHA1       c                   const(2)

     D HASH_SHA256     c                   const(3)

     D HASH_SHA384     c                   const(4)

     D HASH_SHA512     c                   const(5)



     D Qc3DestroyAlgorithmContext...

     D                 PR                  ExtProc('Qc3DestroyAlgorithmContext')

     D   ctxToken                     8a   const

     D   errorCode                         likeds(ERRC0100_t)



     D Qc3CalculateHash...

     D                 PR                  ExtProc('Qc3CalculateHash')

     D   inputData                     *   value

     D   inputDataLen                10i 0 const

     D   inputDataFmt                 8a   const

     D   algDesc                  32767a   const options(*varsize)

     D   algDescFmt                   8a   const

     D   CSP                          1a   const

     D   CSPDev                      10a   const

     D   OutputHash                  64a   options(*varsize:*omit)

     D   errorCode                         likeds(ERRC0100_t)



     D CSP_ANY         c                   const('0')

     D CSP_ANY_DEV     c                   const(' ')



      *

      * MI builtin to create a hex dump of a spot in memory

      *

     D hexdump         PR                  EXTPROC('cvthc')

     D  output                       40A

     D  input                        20A

     D  output_len                   10I 0 value



     D ReportError     PR

     D   errorCode                         likeds(ERRC0100_t) const

     D                                     options(*nopass)



     D showHash        PR

     D   hash                       128a   const



     D fsstats         ds                  likeds(ds_statvfs)

     D ReadBuf         S               *   inz(*null)

     D BufSize         s             10I 0 inz(0)

     D Len             s             10I 0 inz(0)

     D fd              s             10I 0 inz(-1)

     D hashLen         s             10i 0

     D binHash         s             64a

     D hexHash         s            128a



     D errorCode       ds                  likeds(ERRC0100_t)

     D                                     inz(*likeds)



      /free

         *inlr = *on;



         //

         // Prevent crash when some fool calls this without

         // passing parameters.

         //



         if %parms < 2;

           return;

         endif;





         //

         // Determine which type of hash is requested

         //



         select;

         when in_Alg = '*MD5';

           ALGD0500.hashAlg = HASH_MD5;

           hashLen = 16;

         when in_Alg = '*SHA1';

           ALGD0500.hashAlg = HASH_SHA1;

           hashLen = 20;

         when in_Alg = '*SHA256';

           ALGD0500.hashAlg = HASH_SHA256;

           hashLen = 32;

         when in_Alg = '*SHA384';

           ALGD0500.hashAlg = HASH_SHA384;

           hashLen = 48;

         when in_Alg = '*SHA512';

           ALGD0500.hashAlg = HASH_SHA512;

           hashLen = 64;

         other;

           return;

         endsl;





         //

         // Build an algorithm context for hashing

         //

         // NOTE: After this is successful, we must not exit without

         //       first destroying the context (would be a memory

         //       leak)

         //



         Qc3CreateAlgorithmContext( ALGD0500

                                  : 'ALGD0500'

                                  : ALGD0100.ctxToken

                                  : errorCode );

         if errorCode.bytesAvail > 0;

            reportError(errorCode);

         endif;





         //

         //  Get file system statistics to find out the size of a

         //  disk block. IFS Disk access is fastest when you read

         //  an even multiple of these blocks.

         //



         if (statvfs( %trimr(in_Stmf): fsstats ) = -1);

            exsr cleanup;

            reportError();

         endif;





         //

         //  Allocate a buffer for efficient stream file reading

         //

         //  NOTE: After this point, we must not exit without first

         //        deallocating this buffer (would be a memory leak)

         //



         BufSize = fsstats.f_bsize * 8;

         ReadBuf = %alloc(BufSize);





         //

         // Open the stream file for reading

         //



         fd = open( %trimr(in_Stmf) : O_RDONLY );

         if (fd < 0);

            exsr cleanup;

            ReportError();

         endif;





         //

         //  Read stream file, calculate hash on each chunk

         //  of data.

         //



         len = read(fd: readBuf: bufSize);





         dow (len > 0);



             ALGD0100.finalFlag = *OFF;



             Qc3CalculateHash( readBuf

                             : len

                             : 'DATA0100'

                             : ALGD0100

                             : 'ALGD0100'

                             : CSP_ANY

                             : CSP_ANY_DEV

                             : *omit

                             : errorCode );



             if errorCode.bytesAvail > 0;

                exsr Cleanup;

                reportError(errorCode);

             endif;



             len = read(fd: readBuf: bufSize);

         enddo;





         //

         //  At the very end, call API with finalFlag set

         //  to get back the hash

         //



         ALGD0100.finalFlag = *ON;



         Qc3CalculateHash( readBuf

                         : 0

                         : 'DATA0100'

                         : ALGD0100

                         : 'ALGD0100'

                         : CSP_ANY

                         : CSP_ANY_DEV

                         : binHash

                         : errorCode );



         if errorCode.bytesAvail > 0;

            exsr Cleanup;

            reportError(errorCode);

         endif;



         exsr Cleanup;



         //

         //  Apps that use hashes usually want them to be

         //  hexidecimal and lowercase.  Here, that is done...

         //



         hexdump( HexHash : BinHash : hashLen * 2 );

         HexHash = %xlate('ABCDEF': 'abcdef': HexHash);



         //

         // Show the hash on the screen...  note: you may wish

         // to change this to return the hash as a parameter

         // or something to use in a program?

         //



         showHash(HexHash);

         return;





         begsr Cleanup;



           if ALGD0100.ctxToken <> *allx'00';

              Qc3DestroyAlgorithmContext( ALGD0100.ctxToken

                                        : errorCode );

              ALGD0100.ctxToken = *ALLx'00';

           endif;



           if fd <> -1;

              callp close(fd);

              fd = -1;

           endif;



           if readBuf <> *null;

              dealloc readBuf;

              readBuf = *null;

           endif;



        endsr;



      /end-free







      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      * ReportError():  Send an escape message explaining any errors

      *                 that occurred.

      *

      *  This function requires binding directory QC2LE in order

      *  to access the __errno() function.

      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

     P ReportError     B

     D ReportError     PI

     D   errorCode                         likeds(ERRC0100_t) const

     D                                     options(*nopass)



     D get_errno       PR              *   ExtProc('__errno')

     D ptrToErrno      s               *

     D errno           s             10I 0 based(ptrToErrno)



     D QMHSNDPM        PR                  ExtPgm('QMHSNDPM')

     D   MessageID                    7A   Const

     D   QualMsgF                    20A   Const

     D   MsgData                  32767A   Const options(*varsize)

     D   MsgDtaLen                   10I 0 Const

     D   MsgType                     10A   Const

     D   CallStkEnt                  10A   Const

     D   CallStkCnt                  10I 0 Const

     D   MessageKey                   4A

     D   ErrorCode                32767A   options(*varsize)



     D ErrorCrash      DS                  qualified

     D  BytesProv                    10i 0 inz(0)

     D  BytesAvail                   10i 0 inz(0)



     D MsgKey          S              4A

     D MsgID           s              7A

     D MsgDta          s          32767a

     D MsgDtaLen       s             10i 0



      /free



       if %parms >= 1;

          MsgId = errorCode.MsgId;

          MsgDta = errorCode.MsgDta;

          MsgDtaLen = 0;

          if errorCode.bytesAvail > 16;

             MsgDtaLen = errorCode.bytesAvail - 16;

          endif;

       else;

          ptrToErrno = get_errno();

          MsgID = 'CPE' + %char(errno);

          MsgDta = *blanks;

          MsgDtaLen = 0;

       endif;



       QMHSNDPM( MsgID

               : 'QCPFMSG   *LIBL'

               : MsgDta

               : MsgDtaLen

               : '*ESCAPE'

               : '*PGMBDY'

               : 1

               : MsgKey

               : ErrorCrash );



      /end-free

     P                 E







      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      *  ShowHash():  This sends a *COMP program message to display

      *               the calculated hash.

      *

      *   hash = (input) hash to display

      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

     P showHash        B

     D                 PI

     D   hash                       128a   const



     D QMHSNDPM        PR                  ExtPgm('QMHSNDPM')

     D   MessageID                    7A   Const

     D   QualMsgF                    20A   Const

     D   MsgData                  32767A   Const options(*varsize)

     D   MsgDtaLen                   10I 0 Const

     D   MsgType                     10A   Const

     D   CallStkEnt                  10A   Const

     D   CallStkCnt                  10I 0 Const

     D   MessageKey                   4A

     D   ErrorCode                32767A   options(*varsize)



     D ErrorCrash      DS                  qualified

     D  BytesProv                    10i 0 inz(0)

     D  BytesAvail                   10i 0 inz(0)



     D MsgKey          S              4A



      /free



       QMHSNDPM( 'CPF9898'

               : 'QCPFMSG   *LIBL'

               : Hash

               : %len(Hash)

               : '*COMP'

               : '*EXT'

               : 0

               : MsgKey

               : ErrorCrash );



      /end-free

     P                 E

